/**
 * Copyright (c) 2021 Dominic Msters
 * 
 * This software is released under the MIT License.
 * https://opensource.org/licenses/MIT
 */

#include "input.h"

input_t * inputInit(uint32_t inputBindCount, uint32_t inputSourceCount) {
  uint32_t i;

  // Create the input manager
  input_t *input = malloc(sizeof(input_t));
  if(!input) return NULL;

  // Setup the bind lists
  input->inputBindCount = inputBindCount;
  input->bindMap = malloc(sizeof(list_t) * inputBindCount);
  for(i = 0; i < inputBindCount; i++) {
    input->bindMap[i] = listCreate();
  }

  // Create the current & previous input states
  input->current = malloc(sizeof(inputval_t) * inputBindCount);
  input->previous = malloc(sizeof(inputval_t) * inputBindCount);

  // Create the input actuate times array.
  input->times = malloc(sizeof(float) * inputBindCount);

  // Create the buffer, zero all the values out.
  input->buffer = calloc(inputSourceCount, sizeof(inputval_t));

  // Pass off the input manager
  return input;
}

void inputUpdate(input_t *input) {
  uint32_t i;
  listentry_t *item;
  inputval_t value;

  // Flip the states to save memory.
  float * currentCurrent = input->current;
  input->current = input->previous;
  input->previous = currentCurrent;

  // Now look at each bind...
  for(i = 0; i < input->inputBindCount; i++) {
    // Now get the list of input sources bound to this input
    item = input->bindMap[i]->start;
    value = 0;

    // For each input source, add the value from the buffer
    while(item != NULL) {
      value += input->buffer[(inputsource_t)item->data];
      item = item->next;
    }

    // Set to the current state.
    input->current[i] = value;
  }
}

void inputDispose(input_t *input) {
  uint32_t i;

  // Free up the bind lists
  for(i = 0; i < input->inputBindCount; i++) {
    listDispose(input->bindMap[i], false);
  }

  free(input->buffer);
  free(input->times);
  free(input->previous);
  free(input->current);
  free(input->bindMap);
  free(input);
}

void inputBind(input_t *input, inputbind_t bind, inputsource_t source) {
  listAdd(input->bindMap[bind], source);
}

void inputUnbind(input_t *input, inputbind_t bind, inputsource_t source) {
  listRemove(input->bindMap[bind], source);
}

bool inputIsDown(input_t *input, inputbind_t binding) {
  return input->current[binding] != 0;
}

bool inputIsUp(input_t *input, inputbind_t binding) {
  return input->current[binding] == 0;
}

bool inputIsPressed(input_t *input, inputbind_t binding) {
  return (
    input->previous[binding] == 0 &&
    input->current[binding] != 0
  );
}

bool inputIsReleased(input_t *input, inputbind_t binding) {
  return input->current[binding] == 0 && input->previous[binding] != 0;
}

inputval_t inputGetAxis(input_t *input, inputbind_t binding) {
  return input->current[binding];
}

float inputGetFullAxis(input_t *input, inputbind_t positive,
  inputbind_t negative
) {
  return input->current[negative] + input->current[positive];
}

float inputGetAccuated(input_t *input, inputbind_t binding) {
  return input->times[binding];
}