/**
 * Copyright (c) 2021 Dominic Masters
 * 
 * This software is released under the MIT License.
 * https://opensource.org/licenses/MIT
 */

#include "world.h"

world_t * worldLoad(char *assetWorldDirectory) {
  // Define our local variables
  char file [128];
  char *reading, *version, *textureFilename, *temp;
  world_t *world;
  chunk_t *chunk;
  entity_t *entity;
  tiledef_t *tileDef;
  int32_t i, x, y, z;

  // Load the world file.
  file[0] = '\0';
  strcat(file, assetWorldDirectory);
  strcat(file, "world.txt");
  char *worldData = assetStringLoad(file);
  if(worldData == NULL) return NULL;

  // Create the world
  world = malloc(sizeof(world_t));
  world->assetWorldDirectory = assetWorldDirectory;
  world->x = 0, world->y = 0, world->z = 0;

  // Init the entities
  for(i = 0; i < WORLD_ENTITY_COUNT; i++) {
    entity = world->entities + i;
    entity->id = (entityid_t)i;
    entity->entityUpdate = NULL;
    entity->entityDispose = NULL;
  }

  // Now begin parsing, first we need to know which version of the file format
  // we are using.
  reading = worldData;
  version = strtok_r(reading, WORLD_LOAD_TOKEN, &reading);

  // Now load the tileset texture filename
  textureFilename = strtok_r(reading, WORLD_LOAD_TOKEN, &reading);

  // Load tileset texture.
  file[0] = '\0';
  strcat(file, assetWorldDirectory);
  strcat(file, textureFilename);
  world->texture = assetTextureLoad(file);
  
  // Create the tileset
  world->tileset = tilesetCreate(
    world->texture->width/TILE_WIDTH, world->texture->height/TILE_HEIGHT,
    world->texture->width,            world->texture->height,
    0, 0, 0, 0
  );
  
  // Create the Tilemap
  world->tilemap = tileMapCreate(world->tileset->columns, world->tileset->rows);

  // Load Tilemap Definitions. Skip 
  for(i = 0; i < world->tileset->count; i++) {
    tileDef = world->tilemap->tileDefinitions + i;
    
    temp = strtok_r(reading, WORLD_LOAD_TOKEN, &reading);
    tileDef->verticeCount = atoi(temp);

    temp = strtok_r(reading, WORLD_LOAD_TOKEN, &reading);
    tileDef->indiceCount = atoi(temp);

    temp = strtok_r(reading, WORLD_LOAD_TOKEN, &reading);
    tileDef->flags = 0x00;
  }

  // Finished actual loading.
  free(worldData);

  // Load initial chunks, ZYX order is important here.
  i = 0;
  for(z = 0; z < WORLD_DEPTH; z++) {
    for(y = 0; y < WORLD_HEIGHT; y++) {
      for(x = 0; x < WORLD_WIDTH; x++) {
        chunk = world->chunks + i;
        world->chunkList[i] = chunk;

        // Load initial coordinates
        chunk->x = x;
        chunk->y = y;
        chunk->z = z;
        chunk->primitive = NULL;

        // Init the chunk.
        chunkLoad(world, chunk, x, y, z);
        i++;
      }
    }
  }

  return world;
}

void worldRender(world_t *world, shader_t *shader, camera_t *camera) {
  int32_t i, j, tx, ty, tz;
  float x, y, z;
  chunk_t *chunk;
  tile_t *tile;
  tiledef_t *tileDef;
  entity_t *entity;

  // Bind world texture
  shaderUseTexture(shader, world->texture);

  // Render each chunk.
  for(i = 0; i < WORLD_CHUNK_COUNT; i++) {
    chunk = world->chunks + i;
    if(chunk->primitive == NULL) continue;

    x = (chunk->x * CHUNK_WIDTH);
    y = (chunk->y * CHUNK_HEIGHT);
    z = (chunk->z * CHUNK_DEPTH);

    shaderUsePosition(shader, x, y, z, 0, 0, 0);
    primitiveDraw(chunk->primitive, 0, chunk->primitive->indiceCount);
  }

  // Tick entities
  for(i = 0; i < WORLD_ENTITY_COUNT; i++) {
    entity = world->entities + i;
    if(entity->entityUpdate == NULL) break;

    entity->entityUpdate(world, entity, shader, camera);
  }
}

void worldDispose(world_t *world) {
  int32_t i;
  chunk_t *chunk;

  // Unload the chunks
  for(i = 0; i < WORLD_CHUNK_COUNT; i++) {
    chunk = world->chunks + i;
    chunkUnload(world, chunk);
  }

  textureDispose(world->texture);
  tilesetDispose(world->tileset);
  tileMapDispose(world->tilemap);
  free(world);
}

void worldShift(world_t *world, int32_t x, int32_t y, int32_t z) {
  int32_t i,
    /** New List Coordinates */
    lx, ly, lz,
    /** New Local Chunk Coordinates */
    nx, ny, nz,
    /** New Absolute Chunk Coordinates */
    ax, ay, az,
    /** New Chunk Index */
    ni,
    /** Precalculated width * height */
    wh
  ;
  chunk_t *chunk;
  chunk_t *chunkList[WORLD_CHUNK_COUNT];

  // Calculate the new chunklist coordinates
  lx = world->x + x;
  ly = world->y + y;
  lz = world->z + z;

  // Precalc width * height.
  wh = WORLD_WIDTH * WORLD_HEIGHT;

  for(i = 0; i < WORLD_CHUNK_COUNT; i++) {
    chunk = world->chunkList[i];

    // Calculate the new local positions for the chunk.
    nx = mathMod(chunk->x - world->x - x, WORLD_WIDTH);
    ny = mathMod(chunk->y - world->y - y, WORLD_HEIGHT);
    nz = mathMod(chunk->z - world->z - z, WORLD_DEPTH);

    // Load the chunk if we need to. We also use this to calculate new absolutes
    if(
      (ax = lx + nx) != chunk->x ||
      (ly + ny) != chunk->y ||
      (lz + nz) != chunk->z
    ) {
      // Calculate those things that may have not been done within the if
      ay = ly + ny;
      az = lz + nz;
      
      // Load new chunk.
      chunkUnload(world, chunk);
      chunkLoad(world, chunk, ax, ay, az);

      // Update the absolute coordinates.
      chunk->x = ax;
      chunk->y = ay;
      chunk->z = az; 
    }

    // Now, based off those new local positions, calculate the new index.
    ni = (
      nx +
      (ny * WORLD_WIDTH) +
      (nz * wh)
    );
    chunkList[ni] = chunk;
  }

  // Update Absolutes.
  world->x = lx;
  world->y = ly;
  world->z = lz;

  // Now copy that array over.
  memcpy(world->chunkList, chunkList, sizeof(chunk_t *) * WORLD_CHUNK_COUNT);
  free(chunkList);
}

void worldAlign(world_t *world, int32_t x, int32_t y, int32_t z) {
  int32_t lx, ly, lz;
  lx = x - world->x;
  ly = y - world->y;
  lz = z - world->z;
  worldShift(world, lx, ly, lz);
}